/*****************************************************************************/
/*                                                                           */
/*      Version:  1.00   Date: 26/01/96   File: mldoc.c                      */
/* Last Version:                          File:                              */
/* Changes:                                                                  */
/* 26/01/96 Created                                                          */
/*                                                                           */
/* Purpose:                                                                  */
/*                                                                           */
/*                                                                           */
/* Author:  Gerd Stolpmann                                                   */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*===========================================================================*/

#include <stdio.h>
#include <string.h>

#define MAX_ZL 1000
  /* maximale Zeilenlaenge der Eingabedatei */

#define MY_EOF 31

FILE *eingabe;
FILE *ausgabe;

char folgezeile[MAX_ZL];  /* auf `zeile` folgende Zeile */
char zeile[MAX_ZL];/* zu interpretierende Eingabezeile */
int  zp;           /* Index von zeile */
char zc;           /* zu interpretierendes Zeichen */
int  eof_eingabe;  /* wahr, falls die letzte Zeile von 'eingabe' gelesen */
int  letzte_zeile; /* wahr, falls `zeile` die letzte Zeile enthaelt */
int  einrueckung;  /* Einrueckung der Zeile, die auf `zeile` folgt. */
int  zeinrueckung; /* EInrueckung von 'zeile' selbst */
int  zeilennummer; /* Nummer von 'zeile' */

char *trtable[127];
char zeichen[50];


void fehler (char *s)
{
  printf("Zeile %d: %s",zeilennummer,s);
}


char *translation (char c)
{

  if (c==0) {
    return "\n";
  } else if (c=='\t') {
    return " ";
  } else if (c<32) {
    return "";
  } else if (c<127) {
    if (trtable[c]==NULL) {
      zeichen[0]=c;
      zeichen[1]=0;
      return &zeichen;
    } else {
      strcpy(&zeichen,trtable[c]);
      return &zeichen;
    };
  } else
    return "";
}

void init_trtable (void)
{
  int i;

  for (i=0; i<127; i++) trtable[i]=NULL;
  trtable['!']="\\mlchar{33}";
  trtable['"']="\\mlchar{34}";
  trtable['#']="\\mlchar{35}";
  trtable['$']="\\mlchar{36}";
  trtable['%']="\\mlchar{37}";
  trtable['&']="\\mlchar{38}";
  trtable['-']="\\mlchar{45}";
  trtable['<']="\\mlchar{60}";
  trtable['>']="\\mlchar{62}";
  trtable['[']="\\mlchar{91}";
  trtable['\\']="\\mlchar{92}";
  trtable[']']="\\mlchar{93}";
  trtable['^']="\\mlchar{94}";
  trtable['_']="\\mlchar{95}";
  trtable['{']="\\mlchar{123}";
  trtable['|']="\\strich{}";
  trtable['}']="\\mlchar{125}";
  trtable['~']="\\mlchar{126}";
}


void read_next_line (void)
{
  int i;
  int c;

  strcpy(zeile,folgezeile);
  zeinrueckung=einrueckung;
  zeilennummer++;
  /* lies neue Folgezeile ein */
  if (!eof_eingabe) {
    i=0;
    c=getc(eingabe);
    while (c!=EOF && c!='\n' && i<MAX_ZL-1) {
      folgezeile[i]=c;
      c=getc(eingabe);
      i++;
    };
    folgezeile[i]=0;
    eof_eingabe=(c==EOF);
  } else 
      letzte_zeile=1;
  /* Berechne einrueckung */
  i=0;
  einrueckung=0;
  while (folgezeile[i] == ' ' || folgezeile[i] == '\t') {
    if (folgezeile[i] == ' ')
      einrueckung++;
    else
      einrueckung=((einrueckung >> 3)+1) << 3;
    i++;
  };
}


void nextch (void)
{
  if (zc!=MY_EOF) {
    zc=zeile[zp];
    zp++;
    if (zc==0) {
      if (!letzte_zeile) {
	read_next_line();
	zp=0;
      } else
	{
	  zeile[0]=MY_EOF;
	  zp=0;
	};
    };
  };
}


void texdoku_modus (void)
{
  int ende;

  ende=0;
  while (!ende) {
    /* zc=='%' or zc=='|' */
    nextch();
    while (zc==' ' || zc=='\t') nextch();
    while (zc!=0 && zc!='%' && zc!=MY_EOF) {
      fprintf(ausgabe,"%c",zc);
      nextch();
    };
    if (zc=='%') {
      /* Fehler */
      fehler("Dokumentationsteil muss mit % auf Spalte 1 abgeschlossen werden");
      fprintf(ausgabe,"\n");
      ende=1;
    } else {
      /* zc==0 */
      fprintf(ausgabe,"\n");
      /* lies nun das erste Zeichen der Folgezeile */
      nextch();
      if (zc!='%' && zc!='|') {
	/* Fehler */
	fehler("Zeile im Dokumentationsteil beginnt nicht mit | oder %");
      };
      ende=(zc=='%' || zc==MY_EOF);
    };
    fflush(ausgabe);
  };
  /* Das letzte Zeichen, das auf 'ausgabe' geschrieben wurde, ist \n */
  nextch();
}


void mlcode_modus (void)
{
  int within_token, within_string, within_tokenlist, within_char;
  int within_token_b, within_tokenlist_e, within_comment, tex_comment;
  int ist_kommentar_strich;
  char letztes;
  int ende;
  int leerzeichen;
  int indent_erforderlich;
  int indent_gesetzt;
  char *anhaengen;

  fprintf(ausgabe,"\\begin{mlcode}\n");
  ende=0;
  within_token=0;
  within_token_b=0;
  within_string=0;
  within_tokenlist=0;
  within_tokenlist_e=0;
  within_char=0;
  within_comment=0;
  tex_comment=0;
  letztes=0;
  leerzeichen=1;
  indent_erforderlich=0; /*(einrueckung>zeinrueckung);*/
  indent_gesetzt=0;
  anhaengen="";
  while (!ende) {
    /* zc != MY_EOF */
    /* Ggf. ein setindent in den Ausgabestrom einschieben */
    if (indent_erforderlich && zp-1 == einrueckung) {
      fprintf(ausgabe,"\\setindent{}");
      indent_gesetzt=1;
    };
    /* Wechsel zwischen den within-Modi interpretieren */
    if (within_char) {
      if (letztes != '\\' && zc=='\'') within_char=0;
    } else if (within_token_b) {
      if (zc=='`') 
	within_tokenlist=1;
      else
	within_token=1;
      within_token_b=0;
    } else if (within_token) {
      if ((zc=='`' || zc=='\'') && letztes!='\\') within_token=0;
    } else if (within_tokenlist) {
      if (zc=='`' && letztes!='\\') {
	within_tokenlist=0;
	within_tokenlist_e=1;
      }
    } else if (within_tokenlist_e) {
      if (zc=='`') within_tokenlist_e = 0;
	else fehler("Falsche Tokenliste");
    } else if (within_string) {
      if (zc=='"' && letztes!='\\') within_string=0;
    } else if (within_comment) {
      if (letztes=='%' && zc=='-') tex_comment=1;
      if (zc=='%') {
	within_comment=0;
	tex_comment=0;
	anhaengen="\\ecomment{}";
      };
    } else {
      switch (zc) {
      case '%':  { within_comment=1; 
		   fprintf(ausgabe,"\\bcomment{}");
		   break; }
      case '"':  { within_string=1;  break; }
      case '\'': { if (!(isalpha(letztes) || isdigit(letztes) || letztes=='\''
                       || letztes=='_')) within_char=1;
		   break;}
      case '`':  { within_token_b=1; break;}
      };
      /* Endebedingung */
      ende=((zc==';' && letztes==';'));
    };
    ende=ende || (zc==MY_EOF);
    /* Zeichen zc interpretieren */
    ist_kommentar_strich=0;
    if (leerzeichen) {
      if (zc!=' ' && zc!='\t') {
	if (indent_gesetzt) {
	  fprintf(ausgabe,"\\goindent{}");
	  indent_gesetzt=0;
	} else {
	  fprintf(ausgabe,"\\indent{%d}",zeinrueckung);
	};
	leerzeichen=0;
	ist_kommentar_strich=(tex_comment) && (zc=='|');
      };
    };
    if (!leerzeichen) {
      if (tex_comment) {
	if (ist_kommentar_strich)
	  fprintf(ausgabe,"%s%s",translation(zc),anhaengen);
	else
	  fprintf(ausgabe,"%c%s",zc,anhaengen);
      }
      else
	fprintf(ausgabe,"%s%s",translation(zc),anhaengen);
    };
    anhaengen="";
    if (zc==0) {
      indent_erforderlich=0; /*(einrueckung>zeinrueckung);*/
      leerzeichen=1;
      fprintf(ausgabe,"\\nextline{}");
    };
    /* naechstes Zeichen holen */
    letztes=zc;
    nextch();
  };
  if (letztes!=0)
    fprintf(ausgabe,"\n");
  fprintf(ausgabe,"\\end{mlcode}\n");
}



void neutral_modus (void)
{
  int ende;

  ende=0;
  while (!ende) {
    while (zc==' ' || zc=='\t' || zc==0) nextch();
    if (zc==MY_EOF) ende=1;
    else if (zc=='%' && zp==1) {
      texdoku_modus ();
      fprintf(ausgabe,"\\par{}");
    } else {
      mlcode_modus ();
      fprintf(ausgabe,"\\par{}");
    };
  };
}

main(argc,argv)
  int argc;
  char *argv[];
{
  char *mlname, *texname;

  if (argc != 3) exit(1);
  mlname=argv[1];
  texname=argv[2];

  printf("%s\n%s\n",mlname,texname);

  init_trtable();
  eingabe=fopen(mlname, "r");
  ausgabe=fopen(texname, "w");
  eof_eingabe=0;
  letzte_zeile=0;
  folgezeile[0]=0;
  zeilennummer=-1;
  einrueckung=0;
  read_next_line();
  zp=0;

  nextch();
  neutral_modus();

  fclose(eingabe);
  fclose(ausgabe);
}

