##########################################
### Alpha_Equal, Version vom 23-Jan-96 ###
##########################################
;;


%\subsection{Sind zwei Terme $\alpha$-konvertibel?}
|
| Die $\alpha$-Konvertibilit"at von zwei Termen wird mit der Funktion
| {\tt are\_alpha\_equal} festgestellt. Diese erledigt ihre Aufgabe mit
| der zun"achst folgenden rekursiven Funktion {\tt are\_alpha\_equal\_1}.
| Grundgedanke der Funktion ist, die Umbenennungen der gebundenen Variablen
| mitzuvollziehen, um so festzustellen, ob zwei Variable, die in den beiden
| Termen vorkommen, modulo dieser Umbenennung gleich sind. Dazu wird
| rekusiv in alle Teilterme hinabgestiegen, wobei beim Abstieg in einen
| Teilterm festgehalten wird, welche Umbenennungen stattgefunden haben.
| Diese Liste wird dem Teilterm zur Verf"ugung gestellt, der so feststellen
| kann, welche Umbenennungen, die aus irgendeinem Oberterm stammen k"onnen,
| gelten.
|
| {\tt are\_alpha\_equal} hat als Argumente die beiden zu vergleichenden
| Terme {\tt t1} und {\tt t2}, zwei Listen {\tt bound1} und {\tt bound2},
| die f"ur {\tt t1} bzw. {\tt t2} die (in einem Oberterm) gebundenen 
| Variablen enthalten, sowie die Liste {\tt r2}, die die Umbenennungen
| festh"alt. Ist ein Paar {\tt (x,y)} in dieser Liste enthalten, so hat
| die in {\tt t2} vorkommende Variable {\tt x} an den entsprechenden Stellen
| in {\tt t1} den Namen {\tt y}.
|
| Es wird {\tt true} zur"uckgegeben, wenn unter den genannten Voraussetzungen
| die beiden Terme $\alpha$-konvertibel sind.
%
letrec are_alpha_equal_1 (t1:term) (t2:term) 
                         (r2:(var#var)list)
                         (bound1:var list)
                         (bound2:var list) : bool =
  let (opid1, param1), bterms1 = dest_term t1
  and (opid2, param2), bterms2 = dest_term t2 in
    %- Der Term ${\tt t}_i$ hat den Operator ${\tt opid}_i$, die Parameter 
    |  ${\tt param}_i$ und die Liste mit den Teiltermen ${\tt bterms}_i$.
    %
    if (opid1 = opid2) then
    ( %- Bei gleichen Operatoren ist der Fall, da"s es sich um eine Variable
      |  handelt, gesondert zu betrachten.
      %
      if is_variable_term t1 then
      ( %-  Wenn {\tt t1} eine Variable ist, so auch {\tt t2}.
	|   Sie werden als gleich betrachtet, wenn beide frei sind und den
        |   gleichen Namen haben, oder wenn beide gebunden sind und den 
        |   gleichen Namen haben, sofern die mitgef"uhrte Umbenennungsliste
        |   {\tt r2} angewendet wird.
        %
        let v1 = dest_var t1
        and v2 = dest_var t2 in
        let v1bound = mem v1 bound1 
        and v2bound = mem v2 bound2 in
        ( if v1bound & v2bound then
            (v1 = snd(assoc v2 r2) ? v1=v2)
          if (not v1bound) & (not v2bound) then
            v1=v2
          else
	    %- Eine gebundene und eine frei Variable sind immer verschieden. %
            false
        )
      )
      %- Es geht immer noch um den Fall gleicher Operatoren. Sind die Terme
      |  keine Variablen, so sind sie $\alpha$-konvertibel, wenn die 
      |  Parameter gleich sind und der rekursive Abstieg ermittelt, da"s
      |  auch die Teilterme jeweils $\alpha$-konvertibel sind.
      %
      if (param1=param2) then
      ( if length bterms1 = length bterms2 then
          %- {\tt bpairs} ist eine Liste von Paaren, die jeweils die
          |  entsprechenden Teilterme zusammenfassen.
          |  Die Funktion {\tt bound\_alpha\_equal} stellt f"ur zwei
          |  Teilterme {\tt bt1} und {\tt bt2}, die die Variablen 
          |  {\tt bvars1} bzw. {\tt bvars2} als Bindungsvariablen einf"uhren,
          |  ob diese $\alpha$-konvertibel sind.
          %
          let bpairs:((bterm#bterm)list) = combine (bterms1,bterms2) in
          let bound_alpha_equal (bvars1:var list) 
                                (bt1:term)
                                (bvars2:var list)
                                (bt2:term) : bool =
          ( if length bvars1 = length bvars2 then
              let bvpairs = combine (bvars1,bvars2) in
              let difflist =
                itlist (\u1,u2. \diff. if u1=u2 then diff else (u2,u1).diff)
                       bvpairs
                       []
                %- {\tt difflist} ist die lokale Umbenennungsliste %
              in
                are_alpha_equal_1 bt1 bt2 difflist 
                                  (append bound1 bvars1)
                                  (append bound2 bvars2)
            else
              %- Terme, die eine unterschiedliche Anzahl an Bindungsvariablen
              |  einf"uhren, sind nicht $\alpha$-konvertibel.
              %
              false
          )
          in
            %- Nun die $\alpha$-Konvertibilit"at von {\tt t1} und {\tt t2}
            |  feststellen: Diese ist dann gegeben, wenn alle Teilterme
            |  jeweils $\alpha$-konvertibel sind.
            %
            itlist (\(bvars1,bt1),(bvars2,bt2). \zwischen. 
                    zwischen & (bound_alpha_equal bvars1 bt1 bvars2 bt2) )
                   bpairs
                   true
        else
          %- Terme mit verschiedener Anzahl an Teiltermen sind nicht
          |  $\alpha$-konvertibel.
          %
          false
      )
      else
        %- Terme mit verschiedenen Parametern sind nicht $\alpha$-konvertibel.%
        false
    )
    else 
      %- Terme mit unterschiedlichen Operatoren sind nicht
      |  $\alpha$-konvertibel.
      %
      false
;;


% Nun noch die Funktion, die f"ur Anwender gedacht ist und obige Funktion
| aufruft.
%
         
let are_alpha_equal t1 t2 =
  are_alpha_equal_1 t1 t2 [] [] []
;;